"""Provide access to output window and its functionality.

This module provides access to the output window for the currently running
pyRevit command. The proper way to access this wrapper object is through
the :func:`get_output` of :mod:`pyrevit.script` module. This method, in return
uses the `pyrevit.output` module to get access to the output wrapper.

Examples:
    ```python
    from pyrevit import script
    output = script.get_output()
    ```

Here is the source of :func:`pyrevit.script.get_output`. As you can see this
functions calls the :func:`pyrevit.output.get_output` to receive the
output wrapper.

.. literalinclude:: ../../pyrevitlib/pyrevit/script.py
    :pyobject: get_output
"""

from __future__ import print_function
import os.path as op
import sys
from pyrevit.compat import PY2, PY3
if PY2:
    from itertools import izip_longest as zip_longest
elif PY3:
    from itertools import zip_longest

from pyrevit import HOST_APP, EXEC_PARAMS
from pyrevit import framework
from pyrevit import coreutils
from pyrevit.coreutils import logger
from pyrevit.coreutils import markdown, charts
from pyrevit.coreutils import envvars
from pyrevit.runtime.types import ScriptConsoleManager
from pyrevit.output import linkmaker
from pyrevit.userconfig import user_config
from pyrevit import DB

mlogger = logger.get_logger(__name__)

DEFAULT_STYLESHEET_NAME = 'outputstyles.css'


def docclosing_eventhandler(sender, args):
    """Close all output window on document closing."""
    ScriptConsoleManager.CloseActiveOutputWindows()


def setup_output_closer():
    """Setup document closing event listener."""
    HOST_APP.app.DocumentClosing += \
        framework.EventHandler[DB.Events.DocumentClosingEventArgs](
            docclosing_eventhandler
            )


def set_stylesheet(stylesheet):
    """Set active css stylesheet used by output window.

    Args:
        stylesheet (str): full path to stylesheet file
    """
    if op.isfile(stylesheet):
        envvars.set_pyrevit_env_var(envvars.OUTPUT_STYLESHEET_ENVVAR,
                                    stylesheet)
        # do not store this setting forcefully
        # each repo should default to its own stylesheet
        # user_config.output_stylesheet = stylesheet


def get_stylesheet():
    """Return active css stylesheet used by output window."""
    return envvars.get_pyrevit_env_var(envvars.OUTPUT_STYLESHEET_ENVVAR)


def get_default_stylesheet():
    """Return default css stylesheet used by output window."""
    return op.join(op.dirname(__file__), DEFAULT_STYLESHEET_NAME)


def reset_stylesheet():
    """Reset active stylesheet to default."""
    envvars.set_pyrevit_env_var(envvars.OUTPUT_STYLESHEET_ENVVAR,
                                get_default_stylesheet())


# setup output window stylesheet
active_stylesheet = user_config.output_stylesheet or get_default_stylesheet()
set_stylesheet(active_stylesheet)


class PyRevitOutputWindow(object):
    """Wrapper to interact with the output window."""
	
    def __init__(self): 
        self._table_counter = 0
	
    @property
    def window(self):
        """``PyRevitLabs.PyRevit.Runtime.ScriptConsole``: Return output window object."""
        return EXEC_PARAMS.window_handle

    @property
    def renderer(self):
        """Return html renderer inside output window.

        Returns:
            (System.Windows.Forms.WebBrowser): HTML renderer
        """
        if self.window:
            return self.window.renderer

    @property
    def output_id(self):
        """str: Return id of the output window.

        In current implementation, Id of output window is equal to the
        unique id of the pyRevit command it belongs to. This means that all
        output windows belonging to the same pyRevit command, will have
        identical output_id values.
        """
        if self.window:
            return self.window.OutputId

    @property
    def output_uniqueid(self):
        """str: Return unique id of the output window.

        In current implementation, unique id of output window is a GUID string
        generated when the output window is opened. This id is unique to the
        instance of output window.
        """
        if self.window:
            return self.window.OutputUniqueId

    @property
    def is_closed_by_user(self):
        """Whether the window has been closed by the user."""
        return self.window.ClosedByUser

    @property
    def last_line(self):
        """Last line of the output window."""
        return self.window.GetLastLine()

    @property
    def debug_mode(self):
        """Set debug mode on output window and stream.

        This will cause the output window to print information about the
        buffer stream and other aspects of the output window mechanism.
        """
        return EXEC_PARAMS.output_stream.PrintDebugInfo

    @debug_mode.setter
    def debug_mode(self, value):
        EXEC_PARAMS.output_stream.PrintDebugInfo = value

    def _get_head_element(self):
        return self.renderer.Document.GetElementsByTagName('head')[0]

    def _get_body_element(self):
        return self.renderer.Document.GetElementsByTagName('body')[0]

    def self_destruct(self, seconds):
        """Set self-destruct (close window) timer.

        Args:
            seconds (int): number of seconds after which window is closed.
        """
        if self.window:
            self.window.SelfDestructTimer(seconds)

    def inject_to_head(self, element_tag, element_contents, attribs=None):
        """Inject html element to current html head of the output window.

        Args:
            element_tag (str): html tag of the element e.g. 'div'
            element_contents (str): html code of the element contents
            attribs (:obj:`dict`): dictionary of attribute names and value

        Examples:
            ```python
            output = pyrevit.output.get_output()
            output.inject_to_head('script',
                                  '',   # no script since it's a link
                                  {'src': js_script_file_path})
            ```
        """
        html_element = self.renderer.Document.CreateElement(element_tag)
        if element_contents:
            html_element.InnerHtml = element_contents

        if attribs:
            for attribute, value in attribs.items():
                html_element.SetAttribute(attribute, value)

        # inject the script into head
        head_el = self._get_head_element()
        head_el.AppendChild(html_element)
        if self.window:
            self.window.WaitReadyBrowser()

    def inject_to_body(self, element_tag, element_contents, attribs=None):
        """Inject html element to current html body of the output window.

        Args:
            element_tag (str): html tag of the element e.g. 'div'
            element_contents (str): html code of the element contents
            attribs (:obj:`dict`): dictionary of attribute names and value

        Examples:
            ```python
            output = pyrevit.output.get_output()
            output.inject_to_body('script',
                                  '',   # no script since it's a link
                                  {'src': js_script_file_path})
            ```
        """
        html_element = self.renderer.Document.CreateElement(element_tag)
        if element_contents:
            html_element.InnerHtml = element_contents

        if attribs:
            for attribute, value in attribs.items():
                html_element.SetAttribute(attribute, value)

        # inject the script into body
        body_el = self._get_body_element()
        body_el.AppendChild(html_element)
        if self.window:
            self.window.WaitReadyBrowser()

    def inject_script(self, script_code, attribs=None, body=False):
        """Inject script tag into current head (or body) of the output window.

        Args:
            script_code (str): javascript code
            attribs (:obj:`dict`): dictionary of attribute names and value
            body (bool, optional): injects script into body instead of head

        Examples:
            ```python
            output = pyrevit.output.get_output()
            output.inject_script('',   # no script since it's a link
                                 {'src': js_script_file_path})
            ```
        """
        if body:
            self.inject_to_body('script', script_code, attribs=attribs)
        else:
            self.inject_to_head('script', script_code, attribs=attribs)

    def add_style(self, style_code, attribs=None):
        """Inject style tag into current html head of the output window.

        Args:
            style_code (str): css styling code
            attribs (:obj:`dict`): dictionary of attribute names and value

        Examples:
            ```python
            output = pyrevit.output.get_output()
            output.add_style('body { color: blue; }')
            ```
        """
        self.inject_to_head('style', style_code, attribs=attribs)

    def get_head_html(self):
        """str: Return inner code of html head element."""
        return self._get_head_element().InnerHtml

    def set_title(self, new_title):
        """Set window title to the new title."""
        if self.window:
            self.window.Title = new_title

    def set_width(self, width):
        """Set window width to the new width."""
        if self.window:
            self.window.Width = width

    def set_height(self, height):
        """Set window height to the new height."""
        if self.window:
            self.window.Height = height

    def set_font(self, font_family, font_size):
        """Set window font family to the new font family and size.

        Args:
            font_family (str): font family name e.g. 'Courier New'
            font_size (int): font size e.g. 16
        """
        self.renderer.Font = \
            framework.Drawing.Font(font_family,
                                   font_size,
                                   framework.Drawing.FontStyle.Regular,
                                   framework.Drawing.GraphicsUnit.Point)

    def resize(self, width, height):
        """Resize window to the new width and height."""
        self.set_width(width)
        self.set_height(height)

    def center(self):
        """Center the output window on the screen."""
        screen_area = HOST_APP.proc_screen_workarea
        left = \
            (abs(screen_area.Right - screen_area.Left) / 2) \
                - (self.get_width() / 2)
        top = \
            (abs(screen_area.Top - screen_area.Bottom) / 2) \
                - (self.get_height() / 2)
        self.window.Left = left
        self.window.Top = top

    def get_title(self):
        """str: Return current window title."""
        if self.window:
            return self.window.Text

    def get_width(self):
        """int: Return current window width."""
        if self.window:
            return self.window.Width

    def get_height(self):
        """int: Return current window height."""
        if self.window:
            return self.window.Height

    def close(self):
        """Close the window."""
        if self.window:
            self.window.Close()

    def close_others(self, all_open_outputs=False):
        """Close all other windows that belong to the current command.

        Args:
            all_open_outputs (bool): Close all any other windows if True
        """
        if all_open_outputs:
            ScriptConsoleManager.CloseActiveOutputWindows(self.window)
        else:
            ScriptConsoleManager.CloseActiveOutputWindows(self.window,
                                                         self.output_id)

    def hide(self):
        """Hide the window."""
        if self.window:
            self.window.Hide()

    def show(self):
        """Show the window."""
        if self.window:
            self.window.Show()

    def lock_size(self):
        """Lock window size."""
        if self.window:
            self.window.LockSize()

    def unlock_size(self):
        """Unock window size."""
        if self.window:
            self.window.UnlockSize()

    def freeze(self):
        """Freeze output content update."""
        if self.window:
            self.window.Freeze()

    def unfreeze(self):
        """Unfreeze output content update."""
        if self.window:
            self.window.Unfreeze()

    def save_contents(self, dest_file):
        """Save html code of the window.

        Args:
            dest_file (str): full path of the destination html file
        """
        if self.renderer:
            html = \
                self.renderer.Document.Body.OuterHtml.encode('ascii', 'ignore')
            doc_txt = self.renderer.DocumentText
            full_html = doc_txt.lower().replace('<body></body>', html)
            with open(dest_file, 'w') as output_file:
                output_file.write(full_html)

    def open_url(self, dest_url):
        """Open url page in output window.

        Args:
            dest_url (str): web url of the target page
        """
        if self.renderer:
            self.renderer.Navigate(dest_url, False)

    def open_page(self, dest_file):
        """Open html page in output window.

        Args:
            dest_file (str): full path of the target html file
        """
        self.show()
        self.open_url('file:///' + dest_file)

    def update_progress(self, cur_value, max_value):
        """Activate and update the output window progress bar.

        Args:
            cur_value (float): current progress value e.g. 50
            max_value (float): total value e.g. 100

        Examples:
            ```python
            output = pyrevit.output.get_output()
            for i in range(100):
                output.update_progress(i, 100)
            ```
        """
        if self.window:
            self.window.UpdateActivityBar(cur_value, max_value)

    def reset_progress(self):
        """Reset output window progress bar to zero."""
        if self.window:
            self.window.UpdateActivityBar(0, 1)

    def hide_progress(self):
        """Hide output window progress bar."""
        if self.window:
            self.window.SetActivityBarVisibility(False)

    def unhide_progress(self):
        """Unhide output window progress bar."""
        if self.window:
            self.window.SetActivityBarVisibility(True)

    def indeterminate_progress(self, state):
        """Show or hide indeterminate progress bar."""
        if self.window:
            self.window.UpdateActivityBar(state)

    def show_logpanel(self):
        """Show output window logging panel."""
        if self.window:
            self.window.SetActivityBarVisibility(True)

    def hide_logpanel(self):
        """Hide output window logging panel."""
        if self.window:
            self.show_logpanel()
            self.window.SetActivityBarVisibility(False)

    def log_debug(self, message):
        """Report DEBUG message into output logging panel."""
        if self.window:
            self.show_logpanel()
            self.window.activityBar.ConsoleLog(message)

    def log_success(self, message):
        """Report SUCCESS message into output logging panel."""
        if self.window:
            self.show_logpanel()
            self.window.activityBar.ConsoleLogOK(message)

    def log_info(self, message):
        """Report INFO message into output logging panel."""
        if self.window:
            self.show_logpanel()
            self.window.activityBar.ConsoleLogInfo(message)

    def log_warning(self, message):
        """Report WARNING message into output logging panel."""
        if self.window:
            self.show_logpanel()
            self.window.activityBar.ConsoleLogWarning(message)

    def log_error(self, message):
        """Report ERROR message into output logging panel."""
        if self.window:
            self.show_logpanel()
            self.window.activityBar.ConsoleLogError(message)

    def set_icon(self, iconpath):
        """Sets icon on the output window."""
        if self.window:
            self.window.SetIcon(iconpath)

    def reset_icon(self):
        """Sets icon on the output window."""
        if self.window:
            self.window.ResetIcon()

    @staticmethod
    def print_html(html_str):
        """Add the html code to the output window.

        Examples:
            ```python
            output = pyrevit.output.get_output()
            output.print_html('<strong>Title</strong>')
            ```
        """
        print(coreutils.prepare_html_str(html_str),
              end="")

    @staticmethod
    def print_code(code_str):
        """Print code to the output window with special formatting.

        Examples:
            ```python
            output = pyrevit.output.get_output()
            output.print_code('value = 12')
            ```
        """
        code_div = '<div class="code">{}</div>'
        print(
            coreutils.prepare_html_str(
                code_div.format(
                    code_str.replace('    ', '&nbsp;'*4)
                    )
                ),
            end=""
            )

    @staticmethod
    def print_md(md_str):
        """Process markdown code and print to output window.

        Examples:
            ```python
            output = pyrevit.output.get_output()
            output.print_md('### Title')
            ```
        """
        tables_ext = 'pyrevit.coreutils.markdown.extensions.tables'
        markdown_html = markdown.markdown(md_str, extensions=[tables_ext])
        markdown_html = markdown_html.replace('\n', '').replace('\r', '')
        html_code = coreutils.prepare_html_str(markdown_html)
        print(html_code, end="")
        if PY3:
            sys.stdout.flush()


    def print_table(self, table_data, columns=None, formats=None,
                    title='', last_line_style=''):
        """Print provided data in a table in output window.

        Args:
            table_data (list[iterable[Any]]): 2D array of data
            title (str): table title
            columns (list[str]): list of column names
            formats (list[str]): column data formats
            last_line_style (str): css style of last row

        Examples:
            ```python
            data = [
            ['row1', 'data', 'data', 80 ],
            ['row2', 'data', 'data', 45 ],
            ]
            output.print_table(
            table_data=data,
            title="Example Table",
            columns=["Row Name", "Column 1", "Column 2", "Percentage"],
            formats=['', '', '', '{}%'],
            last_line_style='color:red;'
            )
            ```
        """
        if not columns:
            columns = []
        if not formats:
            formats = []

        if last_line_style:
            self.add_style('tr:last-child {{ {style} }}'
                           .format(style=last_line_style))

        zipper = zip_longest #pylint: disable=E1101
        adjust_base_col = '|'
        adjust_extra_col = ':---|'
        base_col = '|'
        extra_col = '{data}|'

        # find max column count
        max_col = max([len(x) for x in table_data])

        header = ''
        if columns:
            header = base_col
            for idx, col_name in zipper(range(max_col), columns, fillvalue=''): #pylint: disable=W0612
                header += extra_col.format(data=col_name)

            header += '\n'

        justifier = adjust_base_col
        for idx in range(max_col):
            justifier += adjust_extra_col

        justifier += '\n'

        rows = ''
        for entry in table_data:
            row = base_col
            for idx, attrib, attr_format \
                    in zipper(range(max_col), entry, formats, fillvalue=''):
                if attr_format:
                    value = attr_format.format(attrib)
                else:
                    value = attrib
                row += extra_col.format(data=value)
            rows += row + '\n'

        table = header + justifier + rows
        self.print_md('### {title}'.format(title=title))
        self.print_md(table)


    def table_html_header(self, columns, table_uid, border_style):
        """Helper method for print_html_table() method

        Return html <thead><tr><th> row for the table header
        
        Args:
            columns (list[str]): list of column names
            table_uid (str): a unique ID for this table's CSS classes
            border_style (str): CSS border style string for table cells
        
        Returns:
            str: HTML string containing the table header row
            
        Examples:
            ```python
            output = pyrevit.output.get_output()
            
            # Basic usage - called internally by print_html_table()
            columns = ["Name", "Age", "City"]
            table_uid = 1
            border_style = "border: 1px solid black;"
            header_html = output.table_html_header(
                columns, table_uid, border_style)
            # Returns: "<thead><tr style='border: 1px solid black;'>" \
            #          "<th class='head_title-1-0' align='left'>Name</th>" \
            #          "<th class='head_title-1-1' align='left'>Age</th>" \
            #          "<th class='head_title-1-2' align='left'>City</th>" \
            #          "</tr></thead>"
            
            # Without border style
            header_html = output.table_html_header(
                columns, table_uid, "")
            # Returns: "<thead><tr>" \
            #          "<th class='head_title-1-0' align='left'>Name</th>" \
            #          "<th class='head_title-1-1' align='left'>Age</th>" \
            #          "<th class='head_title-1-2' align='left'>City</th>" \
            #          "</tr></thead>"
            ```
        """
        html_head = "<thead><tr {}>".format(border_style)
        for i, c in enumerate(columns):
            html_head += \
                "<th class='head_title-{}-{}' align='left'>{}</th>".format(
                    table_uid, i, c)
            # pyRevit original print_table uses align='left'.
            # This is now overridden by CSS if specified
        html_head += "</tr></thead>"

        return html_head


    def table_check_input_lists(self, 
                                table_data,
                                columns,
                                formats,
                                input_kwargs):
        """Helper method for print_html_table() method
        
        Check that the table_data is present and is a list
        Check that table_data rows are of the same length
        Check that all print_html_table() kwargs of type list are of correct length

        Args:
            table_data (list[list[Any]]): The whole table data as 2D list
            columns (list[str]): list of column names
            formats (list[str]): list of format strings for each column
            input_kwargs (list[list[Any]]): list of additional argument lists
        
        Returns:
            tuple: (bool, str) - (True/False, message) indicating result
            
        Examples:
            ```python
            output = pyrevit.output.get_output()
            
            # Valid table data
            table_data = [["John", 25, "NYC"], ["Jane", 30, "LA"]]
            columns = ["Name", "Age", "City"]
            formats = ["", "{} years", ""]
            input_kwargs = [["left", "center", "right"], 
                           ["100px", "80px", "120px"]]
            
            is_valid, message = output.table_check_input_lists(
                table_data, columns, formats, input_kwargs)
            # Returns: (True, "Inputs OK")
            
            # Invalid - mismatched column count
            table_data = [["John", 25], ["Jane", 30, "LA"]]  # Inconsistent
            is_valid, message = output.table_check_input_lists(
                table_data, columns, formats, input_kwargs)
            # Returns: (False, "Not all rows of table_data are of "
            #           "equal length")
            
            # Invalid - wrong number of columns
            columns = ["Name", "Age"]  # Only 2 columns but data has 3
            is_valid, message = output.table_check_input_lists(
                table_data, columns, formats, input_kwargs)
            # Returns: (False, "Column head list length not equal "
            #           "to data row")
            
            # Invalid - empty table data
            is_valid, message = output.table_check_input_lists(
                [], columns, formats, input_kwargs)
            # Returns: (False, "No table_data list")
            ```
        """
  
        # First check positional and named keyword args
        if not table_data:
            return False, "No table_data list"
        if not isinstance(table_data, list):
            return False, "table_data is not a list"
        # table_data is a list. The first sublist must also be a list
        first_data_row = list(table_data[0])
        if not isinstance(first_data_row, list):
            return False, "table_data's first row is not a list or tuple ({})".format(type(first_data_row))
        len_data_row = len(first_data_row)
        if not all(len(row) == len_data_row for row in table_data):
            return False, "Not all rows of table_data are of equal length"
            
        if columns and len_data_row != len(columns): # columns is allowed to be None
            return False, "Column head list length not equal to data row"
        
        if formats and len_data_row != len(formats): # formats is allowed to be None
            return False, "Formats list length not equal to data row"
        
        # Next check **kwargs
        # Loop through the lists and return if not a list or len not equal
        for kwarg_list in input_kwargs:
            if not kwarg_list: # No kwarg is OK beacause they are optional
                continue
            if not isinstance(kwarg_list, list):
                return False, "One of the print_html_table kwargs that should be a list is not a list ({})".format(kwarg_list)
            if len(kwarg_list) != len_data_row:
                return False, "print_html_table kwarg list length problem (should match {} columns)".format(len_data_row)

        return True, "Inputs OK"


    def print_html_table(self,
                    table_data,
                    columns=None,
                    formats=None,
                    title='',
                    last_line_style='',
                    **kwargs):
        """Print provided data in a HTML table in output window.
           The same window can output several tables, each with their own formatting options.

        Args:
            table_data (list[iterable[Any]]): 2D array of data
            title (str): table title
            columns (list[str]): list of column names
            formats (list[str]): column data formats using python string formatting
            last_line_style (str): css style of last row of data (NB applies to all tables in this output)
            column_head_align_styles (list[str]): list css align-text styles for header row
            column_data_align_styles (list[str]): list css align-text styles for data rows
            column_widths (list[str]): list of CSS widths in either px or % 
            column_vertical_border_style (str): CSS compact border definition
            table_width_style (str): CSS to use for width for the whole table, in either px or % 
            repeat_head_as_foot (bool): Repeat the header row at the table foot (useful for long tables)
            row_striping (bool): False to override the default white-grey row stripes and make all white)


        Examples:
            ```python
            data = [
            ['row1', 'data', 'data', 80 ],
            ['row2', 'data', 'data', 45 ],
            ]
            output.print_html_table(
            table_data=data,
            title="Example Table",
            columns=["Row Name", "Column 1", "Column 2", "Percentage"],
            formats=['', '', '', '{}%'],
            last_line_style='color:red;',
            col_head_align_styles = ["left", "left", "center", "right"],
            col_data_align_styles = ["left", "left", "center", "right"],
            column_widths = ["100px", "100px", "500px", "100px"],
            column_vertical_border_style = "border:black solid 1px",
            table_width_style='width:100%', 
            repeat_head_as_foot=True,
            row_striping=False

            )
            Returns:
                Directly prints:
                    print_md of the title (str): 
                    print_html of the generated HTML table with formatting.
            ```
        """

        # Get user formatting options from kwargs
        column_head_align_styles     = kwargs.get("column_head_align_styles", None)
        column_data_align_styles     = kwargs.get("column_data_align_styles", None)
        column_widths                = kwargs.get("column_widths", None)
        column_vertical_border_style = kwargs.get("column_vertical_border_style", None)
        table_width_style            = kwargs.get("table_width_style", None)
        repeat_head_as_foot          = kwargs.get("repeat_head_as_foot", False)
        row_striping                 = kwargs.get("row_striping", True)


        # Get a unique ID for each table from _table_counter
        # This is used in HTML tags to define CSS classes for formatting per table
        self._table_counter += 1
        table_uid = self._table_counter

        # Validate input arguments should be lists:
        input_kwargs = [column_head_align_styles, column_data_align_styles, column_widths]
        inputs_ok, inputs_msg = self.table_check_input_lists(table_data,
                                             columns,
                                             formats,
                                             input_kwargs)

        if not inputs_ok:
            self.print_md('### :warning: {} '.format(inputs_msg))
            return

        
        if not row_striping:
            # Override default original pyRevit white-grey row striping. Makes all rows white.
            self.add_style('tr.data-row-{} {{ {style} }}'.format(table_uid, style="background-color: #ffffff"))

        if last_line_style:
            # Adds a CCS class to allow a last-row format per table (if several in the same output)
            self.add_style('tr.data-row-{}:last-child {{ {style} }}'.format(table_uid, style=last_line_style))

        if column_head_align_styles:
            for i, s in enumerate(column_head_align_styles):
                self.add_style('.head_title-{}-{} {{ text-align:{style} }}'.format(table_uid, i, style=s))

        if column_data_align_styles:
            for i, s in enumerate(column_data_align_styles):
                self.add_style('.data_cell-{}-{} {{ text-align:{style} }}'.format(table_uid, i, style=s))

        if table_width_style:
            self.add_style('.tab-{} {{ width:{} }}'.format(table_uid, table_width_style))

	
        # Open HTML table and its CSS class
        html = "<table class='tab-{}'>".format(table_uid)

        # Build colgroup if user argument specifies column widths
        if column_widths:
              COL = "<col style='width: {}'>"
              html += '<colgroup>'
              for w in column_widths:
                  html += COL.format(w)
              html += "</colgroup>"
        
        if column_vertical_border_style:
            border_style = "style='{}'".format(column_vertical_border_style)
        else:
            border_style = ""

        # Build header row (column titles) if requested
        if columns:
            html_head = self.table_html_header(columns, table_uid, border_style)
            html += html_head
        else:
            html_head =""
            repeat_head_as_foot = False

        
        # Build body rows from 2D python list table_data
        html += "<tbody>"
        for row in table_data:
            # Build an HTML data row with CSS class for this table
            html += "<tr class='data-row-{}'>".format(table_uid)
            for i, cell_data in enumerate(row):
                
                # Slight workaround to be backwards compatible with pyRevit original print_table
                # pyRevit documentation gives the example: formats=['', '', '', '{}%'],
                # Sometimes giving an empty string, sometimes a placeholder with string formatting
                if formats: # If format options provided
                    format_placeholder = formats[i] if formats[i] else "{}"
                    
                    cell_data = format_placeholder.format(cell_data) # Insert data with or without formatting

                html += "<td class='data_cell-{}-{}' {}>{}</td>".format(table_uid, i, border_style, cell_data)
            html += "</tr>"
        
        # Re-insert header row at the end, if requested and if column titles provided
        if repeat_head_as_foot:
            html += html_head


        html += "</tbody>"
        html += "</table>" # Close table

        if title:
            self.print_md('### {title}'.format(title=title))
        self.print_html(html)
    

    def print_image(self, image_path):
        r"""Prints given image to the output.

        Examples:
            ```python
            output = pyrevit.output.get_output()
            output.print_image(r'C:\image.gif')
            ```
        """
        self.print_html(
            "<span><img src=\"file:///{0}\"></span>".format(
                image_path
            )
        )

    def insert_divider(self, level=''):
        """Add horizontal rule to the output window."""
        self.print_md('%s\n-----' % level)

    def next_page(self):
        """Add hidden next page tag to the output window.

        This is helpful to silently separate the output to multiple pages
        for better printing.
        """
        self.print_html('<div class="nextpage"></div><div>&nbsp</div>')

    @staticmethod
    def linkify(element_ids, title=None):
        """Create clickable link for the provided element ids.

        This method, creates the link but does not print it directly.

        Args:
            element_ids (ElementId | list[ElementId]): single or multiple ids
            title (str): tile of the link. defaults to list of element ids
            

        Returns:
            (str): clickable link

        Examples:
            ```python
            output = pyrevit.output.get_output()
            for idx, elid in enumerate(element_ids):
                print('{}: {}'.format(idx+1, output.linkify(elid)))
            ```
        """
        return coreutils.prepare_html_str(
            linkmaker.make_link(element_ids, contents=title)
            )

    def make_chart(self, version=None):
        """:obj:`PyRevitOutputChart`: Return chart object."""
        return charts.PyRevitOutputChart(self, version=version)

    def make_line_chart(self, version=None):
        """:obj:`PyRevitOutputChart`: Return line chart object."""
        return charts.PyRevitOutputChart(
            self,
            chart_type=charts.LINE_CHART,
            version=version
            )

    def make_stacked_chart(self, version=None):
        """:obj:`PyRevitOutputChart`: Return stacked chart object."""
        chart = charts.PyRevitOutputChart(
            self,
            chart_type=charts.LINE_CHART,
            version=version
            )
        chart.options.scales = {'yAxes': [{'stacked': True, }]}
        return chart

    def make_bar_chart(self, version=None):
        """:obj:`PyRevitOutputChart`: Return bar chart object."""
        return charts.PyRevitOutputChart(
            self,
            chart_type=charts.BAR_CHART,
            version=version
            )

    def make_radar_chart(self, version=None):
        """:obj:`PyRevitOutputChart`: Return radar chart object."""
        return charts.PyRevitOutputChart(
            self,
            chart_type=charts.RADAR_CHART,
            version=version
            )

    def make_polar_chart(self, version=None):
        """:obj:`PyRevitOutputChart`: Return polar chart object."""
        return charts.PyRevitOutputChart(
            self,
            chart_type=charts.POLAR_CHART,
            version=version
            )

    def make_pie_chart(self, version=None):
        """:obj:`PyRevitOutputChart`: Return pie chart object."""
        return charts.PyRevitOutputChart(
            self,
            chart_type=charts.PIE_CHART,
            version=version
            )

    def make_doughnut_chart(self, version=None):
        """:obj:`PyRevitOutputChart`: Return dougnut chart object."""
        return charts.PyRevitOutputChart(
            self,
            chart_type=charts.DOUGHNUT_CHART,
            version=version
            )

    def make_bubble_chart(self, version=None):
        """:obj:`PyRevitOutputChart`: Return bubble chart object."""
        return charts.PyRevitOutputChart(
            self,
            chart_type=charts.BUBBLE_CHART,
            version=version
            )


def get_output():
    """:obj:`pyrevit.output.PyRevitOutputWindow` : Return output window."""
    return PyRevitOutputWindow()



class BinaryReader(object,IDisposable):
 """
 Reads primitive data types as binary values in a specific encoding.

 

 BinaryReader(input: Stream)

 BinaryReader(input: Stream,encoding: Encoding)

 BinaryReader(input: Stream,encoding: Encoding,leaveOpen: bool)
 """
 def Close(self):
  """
  Close(self: BinaryReader)

   Closes the current reader and the underlying stream.
  """
  pass
 def Dispose(self):
  """
  Dispose(self: BinaryReader)

   Releases all resources used by the current instance of the System.IO.BinaryReader class.
  """
  pass
 def FillBuffer(self,*args):
  """
  FillBuffer(self: BinaryReader,numBytes: int)

   Fills the internal buffer with the specified number of bytes read from the stream.

  

   numBytes: The number of bytes to be read.
  """
  pass
 def PeekChar(self):
  """
  PeekChar(self: BinaryReader) -> int

  

   Returns the next available character and does not advance the byte or character position.

   Returns: The next available character,or -1 if no more characters are available or the stream does not 

    support seeking.
  """
  pass
 def Read(self,buffer=None,index=None,count=None):
  """
  Read(self: BinaryReader,buffer: Array[Byte],index: int,count: int) -> int

  

   Reads the specified number of bytes from the stream,starting from a specified point in the byte 

    array.

  

  

   buffer: The buffer to read data into.

   index: The starting point in the buffer at which to begin reading into the buffer.

   count: The number of bytes to read.

   Returns: The number of bytes read into buffer. This might be less than the number of bytes requested if 

    that many bytes are not available,or it might be zero if the end of the stream is reached.

  

  Read(self: BinaryReader,buffer: Array[Char],index: int,count: int) -> int

  

   Reads the specified number of characters from the stream,starting from a specified point in the 

    character array.

  

  

   buffer: The buffer to read data into.

   index: The starting point in the buffer at which to begin reading into the buffer.

   count: The number of characters to read.

   Returns: The total number of characters read into the buffer. This might be less than the number of 

    characters requested if that many characters are not currently available,or it might be zero if 

    the end of the stream is reached.

  

  Read(self: BinaryReader) -> int

  

   Reads characters from the underlying stream and advances the current position of the stream in 

    accordance with the Encoding used and the specific character being read from the stream.

  

   Returns: The next character from the input stream,or -1 if no characters are currently available.
  """
  pass
 def Read7BitEncodedInt(self,*args):
  """
  Read7BitEncodedInt(self: BinaryReader) -> int

  

   Reads in a 32-bit integer in compressed format.

   Returns: A 32-bit integer in compressed format.
  """
  pass
 def ReadBoolean(self):
  """
  ReadBoolean(self: BinaryReader) -> bool

  

   Reads a Boolean value from the current stream and advances the current position of the stream by 

    one byte.

  

   Returns: true if the byte is nonzero; otherwise,false.
  """
  pass
 def ReadByte(self):
  """
  ReadByte(self: BinaryReader) -> Byte

  

   Reads the next byte from the current stream and advances the current position of the stream by 

    one byte.

  

   Returns: The next byte read from the current stream.
  """
  pass
 def ReadBytes(self,count):
  """
  ReadBytes(self: BinaryReader,count: int) -> Array[Byte]

  

   Reads the specified number of bytes from the current stream into a byte array and advances the 

    current position by that number of bytes.

  

  

   count: The number of bytes to read.

   Returns: A byte array containing data read from the underlying stream. This might be less than the number 

    of bytes requested if the end of the stream is reached.
  """
  pass
 def ReadChar(self):
  """
  ReadChar(self: BinaryReader) -> Char

  

   Reads the next character from the current stream and advances the current position of the stream 

    in accordance with the Encoding used and the specific character being read from the stream.

  

   Returns: A character read from the current stream.
  """
  pass
 def ReadChars(self,count):
  """
  ReadChars(self: BinaryReader,count: int) -> Array[Char]

  

   Reads the specified number of characters from the current stream,returns the data in a 

    character array,and advances the current position in accordance with the Encoding used and the 

    specific character being read from the stream.

  

  

   count: The number of characters to read.

   Returns: A character array containing data read from the underlying stream. This might be less than the 

    number of characters requested if the end of the stream is reached.
  """
  pass
 def ReadDecimal(self):
  """
  ReadDecimal(self: BinaryReader) -> Decimal

  

   Reads a decimal value from the current stream and advances the current position of the stream by 

    sixteen bytes.

  

   Returns: A decimal value read from the current stream.
  """
  pass
 def ReadDouble(self):
  """
  ReadDouble(self: BinaryReader) -> float

  

   Reads an 8-byte floating point value from the current stream and advances the current position 

    of the stream by eight bytes.

  

   Returns: An 8-byte floating point value read from the current stream.
  """
  pass
 def ReadInt16(self):
  """
  ReadInt16(self: BinaryReader) -> Int16

  

   Reads a 2-byte signed integer from the current stream and advances the current position of the 

    stream by two bytes.

  

   Returns: A 2-byte signed integer read from the current stream.
  """
  pass
 def ReadInt32(self):
  """
  ReadInt32(self: BinaryReader) -> int

  

   Reads a 4-byte signed integer from the current stream and advances the current position of the 

    stream by four bytes.

  

   Returns: A 4-byte signed integer read from the current stream.
  """
  pass
 def ReadInt64(self):
  """
  ReadInt64(self: BinaryReader) -> Int64

  

   Reads an 8-byte signed integer from the current stream and advances the current position of the 

    stream by eight bytes.

  

   Returns: An 8-byte signed integer read from the current stream.
  """
  pass
 def ReadSByte(self):
  """
  ReadSByte(self: BinaryReader) -> SByte

  

   Reads a signed byte from this stream and advances the current position of the stream by one byte.

   Returns: A signed byte read from the current stream.
  """
  pass
 def ReadSingle(self):
  """
  ReadSingle(self: BinaryReader) -> Single

  

   Reads a 4-byte floating point value from the current stream and advances the current position of 

    the stream by four bytes.

  

   Returns: A 4-byte floating point value read from the current stream.
  """
  pass
 def ReadString(self):
  """
  ReadString(self: BinaryReader) -> str

  

   Reads a string from the current stream. The string is prefixed with the length,encoded as an 

    integer seven bits at a time.

  

   Returns: The string being read.
  """
  pass
 def ReadUInt16(self):
  """
  ReadUInt16(self: BinaryReader) -> UInt16

  

   Reads a 2-byte unsigned integer from the current stream using little-endian encoding and 

    advances the position of the stream by two bytes.

  

   Returns: A 2-byte unsigned integer read from this stream.
  """
  pass
 def ReadUInt32(self):
  """
  ReadUInt32(self: BinaryReader) -> UInt32

  

   Reads a 4-byte unsigned integer from the current stream and advances the position of the stream 

    by four bytes.

  

   Returns: A 4-byte unsigned integer read from this stream.
  """
  pass
 def ReadUInt64(self):
  """
  ReadUInt64(self: BinaryReader) -> UInt64

  

   Reads an 8-byte unsigned integer from the current stream and advances the position of the stream 

    by eight bytes.

  

   Returns: An 8-byte unsigned integer read from this stream.
  """
  pass
 def __enter__(self,*args):
  """
  __enter__(self: IDisposable) -> object

  

   Provides the implementation of __enter__ for objects which implement IDisposable.
  """
  pass
 def __exit__(self,*args):
  """
  __exit__(self: IDisposable,exc_type: object,exc_value: object,exc_back: object)

   Provides the implementation of __exit__ for objects which implement IDisposable.
  """
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 @staticmethod
 def __new__(self,input,encoding=None,leaveOpen=None):
  """
  __new__(cls: type,input: Stream)

  __new__(cls: type,input: Stream,encoding: Encoding)

  __new__(cls: type,input: Stream,encoding: Encoding,leaveOpen: bool)
  """
  pass
 def __repr__(self,*args):
  """ __repr__(self: object) -> str """
  pass
 BaseStream=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Exposes access to the underlying stream of the System.IO.BinaryReader.



Get: BaseStream(self: BinaryReader) -> Stream



"""





class TypeDescriptor(object):
 """ Provides information about the characteristics for a component,such as its attributes,properties,and events. This class cannot be inherited. """
 @staticmethod
 def AddAttributes(*__args):
  """
  AddAttributes(instance: object,*attributes: Array[Attribute]) -> TypeDescriptionProvider

  

   Adds class-level attributes to the target component instance.

  

   instance: An instance of the target component.

   attributes: An array of System.Attribute objects to add to the component's class.

   Returns: The newly created System.ComponentModel.TypeDescriptionProvider that was used to add the 

    specified attributes.

  

  AddAttributes(type: Type,*attributes: Array[Attribute]) -> TypeDescriptionProvider

  

   Adds class-level attributes to the target component type.

  

   type: The System.Type of the target component.

   attributes: An array of System.Attribute objects to add to the component's class.

   Returns: The newly created System.ComponentModel.TypeDescriptionProvider that was used to add the 

    specified attributes.
  """
  pass
 @staticmethod
 def AddEditorTable(editorBaseType,table):
  """
  AddEditorTable(editorBaseType: Type,table: Hashtable)

   Adds an editor table for the given editor base type.

  

   editorBaseType: The editor base type to add the editor table for. If a table already exists for this type,this 

    method will do nothing.

  

   table: The System.Collections.Hashtable to add.
  """
  pass
 @staticmethod
 def AddProvider(provider,*__args):
  """
  AddProvider(provider: TypeDescriptionProvider,instance: object)

   Adds a type description provider for a single instance of a component.

  

   provider: The System.ComponentModel.TypeDescriptionProvider to add.

   instance: An instance of the target component.

  AddProvider(provider: TypeDescriptionProvider,type: Type)

   Adds a type description provider for a component class.

  

   provider: The System.ComponentModel.TypeDescriptionProvider to add.

   type: The System.Type of the target component.
  """
  pass
 @staticmethod
 def AddProviderTransparent(provider,*__args):
  """
  AddProviderTransparent(provider: TypeDescriptionProvider,instance: object)

   Adds a type description provider for a single instance of a component.

  

   provider: The System.ComponentModel.TypeDescriptionProvider to add.

   instance: An instance of the target component.

  AddProviderTransparent(provider: TypeDescriptionProvider,type: Type)

   Adds a type description provider for a component class.

  

   provider: The System.ComponentModel.TypeDescriptionProvider to add.

   type: The System.Type of the target component.
  """
  pass
 @staticmethod
 def CreateAssociation(primary,secondary):
  """
  CreateAssociation(primary: object,secondary: object)

   Creates a primary-secondary association between two objects.

  

   primary: The primary System.Object.

   secondary: The secondary System.Object.
  """
  pass
 @staticmethod
 def CreateDesigner(component,designerBaseType):
  """
  CreateDesigner(component: IComponent,designerBaseType: Type) -> IDesigner

  

   Creates an instance of the designer associated with the specified component and of the specified 

    type of designer.

  

  

   component: An System.ComponentModel.IComponent that specifies the component to associate with the designer.

   designerBaseType: A System.Type that represents the type of designer to create.

   Returns: An System.ComponentModel.Design.IDesigner that is an instance of the designer for the component,

    or null if no designer can be found.
  """
  pass
 @staticmethod
 def CreateEvent(componentType,*__args):
  """
  CreateEvent(componentType: Type,oldEventDescriptor: EventDescriptor,*attributes: Array[Attribute]) -> EventDescriptor

  

   Creates a new event descriptor that is identical to an existing event descriptor,when passed 

    the existing System.ComponentModel.EventDescriptor.

  

  

   componentType: The type of the component for which to create the new event.

   oldEventDescriptor: The existing event information.

   attributes: The new attributes.

   Returns: A new System.ComponentModel.EventDescriptor that has merged the specified metadata attributes 

    with the existing metadata attributes.

  

  CreateEvent(componentType: Type,name: str,type: Type,*attributes: Array[Attribute]) -> EventDescriptor

  

   Creates a new event descriptor that is identical to an existing event descriptor by dynamically 

    generating descriptor information from a specified event on a type.

  

  

   componentType: The type of the component the event lives on.

   name: The name of the event.

   type: The type of the delegate that handles the event.

   attributes: The attributes for this event.

   Returns: An System.ComponentModel.EventDescriptor that is bound to a type.
  """
  pass
 @staticmethod
 def CreateInstance(provider,objectType,argTypes,args):
  """
  CreateInstance(provider: IServiceProvider,objectType: Type,argTypes: Array[Type],args: Array[object]) -> object

  

   Creates an object that can substitute for another data type.

  

   provider: The service provider that provides a System.ComponentModel.TypeDescriptionProvider service. This 

    parameter can be null.

  

   objectType: The System.Type of object to create.

   argTypes: An optional array of parameter types to be passed to the object's constructor. This parameter 

    can be null or an array of zero length.

  

   args: An optional array of parameter values to pass to the object's constructor. If not null,the 

    number of elements must be the same as argTypes.

  

   Returns: An instance of the substitute data type if an associated 

    System.ComponentModel.TypeDescriptionProvider is found; otherwise,null.
  """
  pass
 @staticmethod
 def CreateProperty(componentType,*__args):
  """
  CreateProperty(componentType: Type,oldPropertyDescriptor: PropertyDescriptor,*attributes: Array[Attribute]) -> PropertyDescriptor

  

   Creates a new property descriptor from an existing property descriptor,using the specified 

    existing System.ComponentModel.PropertyDescriptor and attribute array.

  

  

   componentType: The System.Type of the component that the property is a member of.

   oldPropertyDescriptor: The existing property descriptor.

   attributes: The new attributes for this property.

   Returns: A new System.ComponentModel.PropertyDescriptor that has the specified metadata attributes merged 

    with the existing metadata attributes.

  

  CreateProperty(componentType: Type,name: str,type: Type,*attributes: Array[Attribute]) -> PropertyDescriptor

  

   Creates and dynamically binds a property descriptor to a type,using the specified property 

    name,type,and attribute array.

  

  

   componentType: The System.Type of the component that the property is a member of.

   name: The name of the property.

   type: The System.Type of the property.

   attributes: The new attributes for this property.

   Returns: A System.ComponentModel.PropertyDescriptor that is bound to the specified type and that has the 

    specified metadata attributes merged with the existing metadata attributes.
  """
  pass
 @staticmethod
 def GetAssociation(type,primary):
  """
  GetAssociation(type: Type,primary: object) -> object

  

   Returns an instance of the type associated with the specified primary object.

  

   type: The System.Type of the target component.

   primary: The primary object of the association.

   Returns: An instance of the secondary type that has been associated with the primary object if an 

    association exists; otherwise,primary if no specified association exists.
  """
  pass
 @staticmethod
 def GetAttributes(*__args):
  """
  GetAttributes(component: object,noCustomTypeDesc: bool) -> AttributeCollection

  

   Returns a collection of attributes for the specified component and a Boolean indicating that a 

    custom type descriptor has been created.

  

  

   component: The component for which you want to get attributes.

   noCustomTypeDesc: true to use a baseline set of attributes from the custom type descriptor if component is of type 

    System.ComponentModel.ICustomTypeDescriptor; otherwise,false.

  

   Returns: An System.ComponentModel.AttributeCollection with the attributes for the component. If the 

    component is null,this method returns an empty collection.

  

  GetAttributes(component: object) -> AttributeCollection

  

   Returns the collection of attributes for the specified component.

  

   component: The component for which you want to get attributes.

   Returns: An System.ComponentModel.AttributeCollection containing the attributes for the component. If 

    component is null,this method returns an empty collection.

  

  GetAttributes(componentType: Type) -> AttributeCollection

  

   Returns a collection of attributes for the specified type of component.

  

   componentType: The System.Type of the target component.

   Returns: An System.ComponentModel.AttributeCollection with the attributes for the type of the component. 

    If the component is null,this method returns an empty collection.
  """
  pass
 @staticmethod
 def GetClassName(*__args):
  """
  GetClassName(componentType: Type) -> str

  

   Returns the name of the class for the specified type.

  

   componentType: The System.Type of the target component.

   Returns: A System.String containing the name of the class for the specified component type.

  GetClassName(component: object,noCustomTypeDesc: bool) -> str

  

   Returns the name of the class for the specified component using a custom type descriptor.

  

   component: The System.Object for which you want the class name.

   noCustomTypeDesc: true to consider custom type description information; otherwise,false.

   Returns: A System.String containing the name of the class for the specified component.

  GetClassName(component: object) -> str

  

   Returns the name of the class for the specified component using the default type descriptor.

  

   component: The System.Object for which you want the class name.

   Returns: A System.String containing the name of the class for the specified component.
  """
  pass
 @staticmethod
 def GetComponentName(component,noCustomTypeDesc=None):
  """
  GetComponentName(component: object,noCustomTypeDesc: bool) -> str

  

   Returns the name of the specified component using a custom type descriptor.

  

   component: The System.Object for which you want the class name.

   noCustomTypeDesc: true to consider custom type description information; otherwise,false.

   Returns: The name of the class for the specified component,or null if there is no component name.

  GetComponentName(component: object) -> str

  

   Returns the name of the specified component using the default type descriptor.

  

   component: The System.Object for which you want the class name.

   Returns: A System.String containing the name of the specified component,or null if there is no component 

    name.
  """
  pass
 @staticmethod
 def GetConverter(*__args):
  """
  GetConverter(type: Type) -> TypeConverter

  

   Returns a type converter for the specified type.

  

   type: The System.Type of the target component.

   Returns: A System.ComponentModel.TypeConverter for the specified type.

  GetConverter(component: object,noCustomTypeDesc: bool) -> TypeConverter

  

   Returns a type converter for the type of the specified component with a custom type descriptor.

  

   component: A component to get the converter for.

   noCustomTypeDesc: true to consider custom type description information; otherwise,false.

   Returns: A System.ComponentModel.TypeConverter for the specified component.

  GetConverter(component: object) -> TypeConverter

  

   Returns a type converter for the type of the specified component.

  

   component: A component to get the converter for.

   Returns: A System.ComponentModel.TypeConverter for the specified component.
  """
  pass
 @staticmethod
 def GetDefaultEvent(*__args):
  """
  GetDefaultEvent(component: object,noCustomTypeDesc: bool) -> EventDescriptor

  

   Returns the default event for a component with a custom type descriptor.

  

   component: The component to get the event for.

   noCustomTypeDesc: true to consider custom type description information; otherwise,false.

   Returns: An System.ComponentModel.EventDescriptor with the default event,or null if there are no events.

  GetDefaultEvent(component: object) -> EventDescriptor

  

   Returns the default event for the specified component.

  

   component: The component to get the event for.

   Returns: An System.ComponentModel.EventDescriptor with the default event,or null if there are no events.

  GetDefaultEvent(componentType: Type) -> EventDescriptor

  

   Returns the default event for the specified type of component.

  

   componentType: The System.Type of the target component.

   Returns: An System.ComponentModel.EventDescriptor with the default event,or null if there are no events.
  """
  pass
 @staticmethod
 def GetDefaultProperty(*__args):
  """
  GetDefaultProperty(component: object,noCustomTypeDesc: bool) -> PropertyDescriptor

  

   Returns the default property for the specified component with a custom type descriptor.

  

   component: The component to get the default property for.

   noCustomTypeDesc: true to consider custom type description information; otherwise,false.

   Returns: A System.ComponentModel.PropertyDescriptor with the default property,or null if there are no 

    properties.

  

  GetDefaultProperty(component: object) -> PropertyDescriptor

  

   Returns the default property for the specified component.

  

   component: The component to get the default property for.

   Returns: A System.ComponentModel.PropertyDescriptor with the default property,or null if there are no 

    properties.

  

  GetDefaultProperty(componentType: Type) -> PropertyDescriptor

  

   Returns the default property for the specified type of component.

  

   componentType: A System.Type that represents the class to get the property for.

   Returns: A System.ComponentModel.PropertyDescriptor with the default property,or null if there are no 

    properties.
  """
  pass
 @staticmethod
 def GetEditor(*__args):
  """
  GetEditor(type: Type,editorBaseType: Type) -> object

  

   Returns an editor with the specified base type for the specified type.

  

   type: The System.Type of the target component.

   editorBaseType: A System.Type that represents the base type of the editor you are trying to find.

   Returns: An instance of the editor object that can be cast to the given base type,or null if no editor 

    of the requested type can be found.

  

  GetEditor(component: object,editorBaseType: Type,noCustomTypeDesc: bool) -> object

  

   Returns an editor with the specified base type and with a custom type descriptor for the 

    specified component.

  

  

   component: The component to get the editor for.

   editorBaseType: A System.Type that represents the base type of the editor you want to find.

   noCustomTypeDesc: A flag indicating whether custom type description information should be considered.

   Returns: An instance of the editor that can be cast to the specified editor type,or null if no editor of 

    the requested type can be found.

  

  GetEditor(component: object,editorBaseType: Type) -> object

  

   Gets an editor with the specified base type for the specified component.

  

   component: The component to get the editor for.

   editorBaseType: A System.Type that represents the base type of the editor you want to find.

   Returns: An instance of the editor that can be cast to the specified editor type,or null if no editor of 

    the requested type can be found.
  """
  pass
 @staticmethod
 def GetEvents(*__args):
  """
  GetEvents(component: object,noCustomTypeDesc: bool) -> EventDescriptorCollection

  

   Returns the collection of events for a specified component with a custom type descriptor.

  

   component: A component to get the events for.

   noCustomTypeDesc: true to consider custom type description information; otherwise,false.

   Returns: An System.ComponentModel.EventDescriptorCollection with the events for this component.

  GetEvents(component: object,attributes: Array[Attribute]) -> EventDescriptorCollection

  

   Returns the collection of events for a specified component using a specified array of attributes 

    as a filter.

  

  

   component: A component to get the events for.

   attributes: An array of type System.Attribute that you can use as a filter.

   Returns: An System.ComponentModel.EventDescriptorCollection with the events that match the specified 

    attributes for this component.

  

  GetEvents(component: object,attributes: Array[Attribute],noCustomTypeDesc: bool) -> EventDescriptorCollection

  

   Returns the collection of events for a specified component using a specified array of attributes 

    as a filter and using a custom type descriptor.

  

  

   component: A component to get the events for.

   attributes: An array of type System.Attribute to use as a filter.

   noCustomTypeDesc: true to consider custom type description information; otherwise,false.

   Returns: An System.ComponentModel.EventDescriptorCollection with the events that match the specified 

    attributes for this component.

  

  GetEvents(componentType: Type) -> EventDescriptorCollection

  

   Returns the collection of events for a specified type of component.

  

   componentType: The System.Type of the target component.

   Returns: An System.ComponentModel.EventDescriptorCollection with the events for this component.

  GetEvents(componentType: Type,attributes: Array[Attribute]) -> EventDescriptorCollection

  

   Returns the collection of events for a specified type of component using a specified array of 

    attributes as a filter.

  

  

   componentType: The System.Type of the target component.

   attributes: An array of type System.Attribute that you can use as a filter.

   Returns: An System.ComponentModel.EventDescriptorCollection with the events that match the specified 

    attributes for this component.

  

  GetEvents(component: object) -> EventDescriptorCollection

  

   Returns the collection of events for the specified component.

  

   component: A component to get the events for.

   Returns: An System.ComponentModel.EventDescriptorCollection with the events for this component.
  """
  pass
 @staticmethod
 def GetFullComponentName(component):
  """
  GetFullComponentName(component: object) -> str

  

   Returns the fully qualified name of the component.

  

   component: The System.ComponentModel.Component to find the name for.

   Returns: The fully qualified name of the specified component,or null if the component has no name.
  """
  pass
 @staticmethod
 def GetProperties(*__args):
  """
  GetProperties(component: object,noCustomTypeDesc: bool) -> PropertyDescriptorCollection

  

   Returns the collection of properties for a specified component using the default type descriptor.

  

   component: A component to get the properties for.

   noCustomTypeDesc: true to not consider custom type description information; otherwise,false.

   Returns: A System.ComponentModel.PropertyDescriptorCollection with the properties for a specified 

    component.

  

  GetProperties(component: object,attributes: Array[Attribute]) -> PropertyDescriptorCollection

  

   Returns the collection of properties for a specified component using a specified array of 

    attributes as a filter.

  

  

   component: A component to get the properties for.

   attributes: An array of type System.Attribute to use as a filter.

   Returns: A System.ComponentModel.PropertyDescriptorCollection with the properties that match the 

    specified attributes for the specified component.

  

  GetProperties(component: object,attributes: Array[Attribute],noCustomTypeDesc: bool) -> PropertyDescriptorCollection

  

   Returns the collection of properties for a specified component using a specified array of 

    attributes as a filter and using a custom type descriptor.

  

  

   component: A component to get the properties for.

   attributes: An array of type System.Attribute to use as a filter.

   noCustomTypeDesc: true to consider custom type description information; otherwise,false.

   Returns: A System.ComponentModel.PropertyDescriptorCollection with the events that match the specified 

    attributes for the specified component.

  

  GetProperties(componentType: Type) -> PropertyDescriptorCollection

  

   Returns the collection of properties for a specified type of component.

  

   componentType: A System.Type that represents the component to get properties for.

   Returns: A System.ComponentModel.PropertyDescriptorCollection with the properties for a specified type of 

    component.

  

  GetProperties(componentType: Type,attributes: Array[Attribute]) -> PropertyDescriptorCollection

  

   Returns the collection of properties for a specified type of component using a specified array 

    of attributes as a filter.

  

  

   componentType: The System.Type of the target component.

   attributes: An array of type System.Attribute to use as a filter.

   Returns: A System.ComponentModel.PropertyDescriptorCollection with the properties that match the 

    specified attributes for this type of component.

  

  GetProperties(component: object) -> PropertyDescriptorCollection

  

   Returns the collection of properties for a specified component.

  

   component: A component to get the properties for.

   Returns: A System.ComponentModel.PropertyDescriptorCollection with the properties for the specified 

    component.
  """
  pass
 @staticmethod
 def GetProvider(*__args):
  """
  GetProvider(instance: object) -> TypeDescriptionProvider

  

   Returns the type description provider for the specified component.

  

   instance: An instance of the target component.

   Returns: A System.ComponentModel.TypeDescriptionProvider associated with the specified component.

  GetProvider(type: Type) -> TypeDescriptionProvider

  

   Returns the type description provider for the specified type.

  

   type: The System.Type of the target component.

   Returns: A System.ComponentModel.TypeDescriptionProvider associated with the specified type.
  """
  pass
 @staticmethod
 def GetReflectionType(*__args):
  """
  GetReflectionType(instance: object) -> Type

  

   Returns a System.Type that can be used to perform reflection,given an object.

  

   instance: An instance of the target component.

   Returns: A System.Type for the specified object.

  GetReflectionType(type: Type) -> Type

  

   Returns a System.Type that can be used to perform reflection,given a class type.

  

   type: The System.Type of the target component.

   Returns: A System.Type of the specified class.
  """
  pass
 @staticmethod
 def Refresh(*__args):
  """
  Refresh(module: Module)

   Clears the properties and events for the specified module from the cache.

  

   module: The System.Reflection.Module that represents the module to refresh. Each System.Type in this 

    module will be refreshed.

  

  Refresh(assembly: Assembly)

   Clears the properties and events for the specified assembly from the cache.

  

   assembly: The System.Reflection.Assembly that represents the assembly to refresh. Each System.Type in this 

    assembly will be refreshed.

  

  Refresh(component: object)

   Clears the properties and events for the specified component from the cache.

  

   component: A component for which the properties or events have changed.

  Refresh(type: Type)

   Clears the properties and events for the specified type of component from the cache.

  

   type: The System.Type of the target component.
  """
  pass
 @staticmethod
 def RemoveAssociation(primary,secondary):
  """
  RemoveAssociation(primary: object,secondary: object)

   Removes an association between two objects.

  

   primary: The primary System.Object.

   secondary: The secondary System.Object.
  """
  pass
 @staticmethod
 def RemoveAssociations(primary):
  """
  RemoveAssociations(primary: object)

   Removes all associations for a primary object.

  

   primary: The primary System.Object in an association.
  """
  pass
 @staticmethod
 def RemoveProvider(provider,*__args):
  """
  RemoveProvider(provider: TypeDescriptionProvider,instance: object)

   Removes a previously added type description provider that is associated with the specified 

    object.

  

  

   provider: The System.ComponentModel.TypeDescriptionProvider to remove.

   instance: An instance of the target component.

  RemoveProvider(provider: TypeDescriptionProvider,type: Type)

   Removes a previously added type description provider that is associated with the specified type.

  

   provider: The System.ComponentModel.TypeDescriptionProvider to remove.

   type: The System.Type of the target component.
  """
  pass
 @staticmethod
 def RemoveProviderTransparent(provider,*__args):
  """
  RemoveProviderTransparent(provider: TypeDescriptionProvider,instance: object)

   Removes a previously added type description provider that is associated with the specified 

    object.

  

  

   provider: The System.ComponentModel.TypeDescriptionProvider to remove.

   instance: An instance of the target component.

  RemoveProviderTransparent(provider: TypeDescriptionProvider,type: Type)

   Removes a previously added type description provider that is associated with the specified type.

  

   provider: The System.ComponentModel.TypeDescriptionProvider to remove.

   type: The System.Type of the target component.
  """
  pass
 @staticmethod
 def SortDescriptorArray(infos):
  """
  SortDescriptorArray(infos: IList)

   Sorts descriptors using the name of the descriptor.

  

   infos: An System.Collections.IList that contains the descriptors to sort.
  """
  pass
 ComNativeDescriptorHandler=None
 ComObjectType=None
 InterfaceType=None
 Refreshed=None




class PropertyDescriptor(MemberDescriptor):
 """ Provides an abstraction of a property on a class. """
 def AddValueChanged(self,component,handler):
  """
  AddValueChanged(self: PropertyDescriptor,component: object,handler: EventHandler)

   Enables other objects to be notified when this property changes.

  

   component: The component to add the handler for.

   handler: The delegate to add as a listener.
  """
  pass
 def CanResetValue(self,component):
  """
  CanResetValue(self: PropertyDescriptor,component: object) -> bool

  

   When overridden in a derived class,returns whether resetting an object changes its value.

  

   component: The component to test for reset capability.

   Returns: true if resetting the component changes its value; otherwise,false.
  """
  pass
 def CreateInstance(self,*args):
  """
  CreateInstance(self: PropertyDescriptor,type: Type) -> object

  

   Creates an instance of the specified type.

  

   type: A System.Type that represents the type to create.

   Returns: A new instance of the type.
  """
  pass
 def Equals(self,obj):
  """
  Equals(self: PropertyDescriptor,obj: object) -> bool

  

   Compares this to another object to see if they are equivalent.

  

   obj: The object to compare to this System.ComponentModel.PropertyDescriptor.

   Returns: true if the values are equivalent; otherwise,false.
  """
  pass
 def GetChildProperties(self,*__args):
  """
  GetChildProperties(self: PropertyDescriptor,instance: object) -> PropertyDescriptorCollection

  

   Returns a System.ComponentModel.PropertyDescriptorCollection for a given object.

  

   instance: A component to get the properties for.

   Returns: A System.ComponentModel.PropertyDescriptorCollection with the properties for the specified 

    component.

  

  GetChildProperties(self: PropertyDescriptor,instance: object,filter: Array[Attribute]) -> PropertyDescriptorCollection

  

   Returns a System.ComponentModel.PropertyDescriptorCollection for a given object using a 

    specified array of attributes as a filter.

  

  

   instance: A component to get the properties for.

   filter: An array of type System.Attribute to use as a filter.

   Returns: A System.ComponentModel.PropertyDescriptorCollection with the properties that match the 

    specified attributes for the specified component.

  

  GetChildProperties(self: PropertyDescriptor) -> PropertyDescriptorCollection

  

   Returns the default System.ComponentModel.PropertyDescriptorCollection.

   Returns: A System.ComponentModel.PropertyDescriptorCollection.

  GetChildProperties(self: PropertyDescriptor,filter: Array[Attribute]) -> PropertyDescriptorCollection

  

   Returns a System.ComponentModel.PropertyDescriptorCollection using a specified array of 

    attributes as a filter.

  

  

   filter: An array of type System.Attribute to use as a filter.

   Returns: A System.ComponentModel.PropertyDescriptorCollection with the properties that match the 

    specified attributes.
  """
  pass
 def GetEditor(self,editorBaseType):
  """
  GetEditor(self: PropertyDescriptor,editorBaseType: Type) -> object

  

   Gets an editor of the specified type.

  

   editorBaseType: The base type of editor,which is used to differentiate between multiple editors that a property 

    supports.

  

   Returns: An instance of the requested editor type,or null if an editor cannot be found.
  """
  pass
 def GetHashCode(self):
  """
  GetHashCode(self: PropertyDescriptor) -> int

  

   Returns the hash code for this object.

   Returns: The hash code for this object.
  """
  pass
 def GetTypeFromName(self,*args):
  """
  GetTypeFromName(self: PropertyDescriptor,typeName: str) -> Type

  

   Returns a type using its name.

  

   typeName: The assembly-qualified name of the type to retrieve.

   Returns: A System.Type that matches the given type name,or null if a match cannot be found.
  """
  pass
 def GetValue(self,component):
  """
  GetValue(self: PropertyDescriptor,component: object) -> object

  

   When overridden in a derived class,gets the current value of the property on a component.

  

   component: The component with the property for which to retrieve the value.

   Returns: The value of a property for a given component.
  """
  pass
 def GetValueChangedHandler(self,*args):
  """
  GetValueChangedHandler(self: PropertyDescriptor,component: object) -> EventHandler

  

   Retrieves the current set of ValueChanged event handlers for a specific component

  

   component: The component for which to retrieve event handlers.

   Returns: A combined multicast event handler,or null if no event handlers are currently assigned to 

    component.
  """
  pass
 def OnValueChanged(self,*args):
  """
  OnValueChanged(self: PropertyDescriptor,component: object,e: EventArgs)

   Raises the ValueChanged event that you implemented.

  

   component: The object that raises the event.

   e: An System.EventArgs that contains the event data.
  """
  pass
 def RemoveValueChanged(self,component,handler):
  """
  RemoveValueChanged(self: PropertyDescriptor,component: object,handler: EventHandler)

   Enables other objects to be notified when this property changes.

  

   component: The component to remove the handler for.

   handler: The delegate to remove as a listener.
  """
  pass
 def ResetValue(self,component):
  """
  ResetValue(self: PropertyDescriptor,component: object)

   When overridden in a derived class,resets the value for this property of the component to the 

    default value.

  

  

   component: The component with the property value that is to be reset to the default value.
  """
  pass
 def SetValue(self,component,value):
  """
  SetValue(self: PropertyDescriptor,component: object,value: object)

   When overridden in a derived class,sets the value of the component to a different value.

  

   component: The component with the property value that is to be set.

   value: The new value.
  """
  pass
 def ShouldSerializeValue(self,component):
  """
  ShouldSerializeValue(self: PropertyDescriptor,component: object) -> bool

  

   When overridden in a derived class,determines a value indicating whether the value of this 

    property needs to be persisted.

  

  

   component: The component with the property to be examined for persistence.

   Returns: true if the property should be persisted; otherwise,false.
  """
  pass
 @staticmethod
 def __new__(self,*args): #cannot find CLR constructor
  """
  __new__(cls: type,name: str,attrs: Array[Attribute])

  __new__(cls: type,descr: MemberDescriptor)

  __new__(cls: type,descr: MemberDescriptor,attrs: Array[Attribute])
  """
  pass
 AttributeArray=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets an array of attributes.



"""

 ComponentType=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """When overridden in a derived class,gets the type of the component this property is bound to.



Get: ComponentType(self: PropertyDescriptor) -> Type



"""

 Converter=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the type converter for this property.



Get: Converter(self: PropertyDescriptor) -> TypeConverter



"""

 IsLocalizable=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value indicating whether this property should be localized,as specified in the System.ComponentModel.LocalizableAttribute.



Get: IsLocalizable(self: PropertyDescriptor) -> bool



"""

 IsReadOnly=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """When overridden in a derived class,gets a value indicating whether this property is read-only.



Get: IsReadOnly(self: PropertyDescriptor) -> bool



"""

 NameHashCode=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the hash code for the name of the member,as specified in System.String.GetHashCode.



"""

 PropertyType=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """When overridden in a derived class,gets the type of the property.



Get: PropertyType(self: PropertyDescriptor) -> Type



"""

 SerializationVisibility=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value indicating whether this property should be serialized,as specified in the System.ComponentModel.DesignerSerializationVisibilityAttribute.



Get: SerializationVisibility(self: PropertyDescriptor) -> DesignerSerializationVisibility



"""

 SupportsChangeEvents=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value indicating whether value change notifications for this property may originate from outside the property descriptor.



Get: SupportsChangeEvents(self: PropertyDescriptor) -> bool



"""





class DesignerSerializationVisibilityAttribute(Attribute,_Attribute):
 """
 Specifies the type of persistence to use when serializing a property on a component at design time.

 

 DesignerSerializationVisibilityAttribute(visibility: DesignerSerializationVisibility)
 """
 def Equals(self,obj):
  """
  Equals(self: DesignerSerializationVisibilityAttribute,obj: object) -> bool

  

   Indicates whether this instance and a specified object are equal.

  

   obj: Another object to compare to.

   Returns: true if obj is equal to this instance; otherwise,false.
  """
  pass
 def GetHashCode(self):
  """
  GetHashCode(self: DesignerSerializationVisibilityAttribute) -> int

  

   Returns the hash code for this object.

   Returns: A 32-bit signed integer hash code.
  """
  pass
 def IsDefaultAttribute(self):
  """
  IsDefaultAttribute(self: DesignerSerializationVisibilityAttribute) -> bool

  

   Gets a value indicating whether the current value of the attribute is the default value for the 

    attribute.

  

   Returns: true if the attribute is set to the default value; otherwise,false.
  """
  pass
 def __eq__(self,*args):
  """ x.__eq__(y) <==> x==y """
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 @staticmethod
 def __new__(self,visibility):
  """ __new__(cls: type,visibility: DesignerSerializationVisibility) """
  pass
 def __ne__(self,*args):
  pass
 Visibility=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value indicating the basic serialization mode a serializer should use when determining whether and how to persist the value of a property.



Get: Visibility(self: DesignerSerializationVisibilityAttribute) -> DesignerSerializationVisibility



"""


 Content=None
 Default=None
 Hidden=None
 Visible=None




class CategoryAttribute(Attribute,_Attribute):
 """
 Specifies the name of the category in which to group the property or event when displayed in a System.Windows.Forms.PropertyGrid control set to Categorized mode.

 

 CategoryAttribute()

 CategoryAttribute(category: str)
 """
 def Equals(self,obj):
  """
  Equals(self: CategoryAttribute,obj: object) -> bool

  

   Returns whether the value of the given object is equal to the current 

    System.ComponentModel.CategoryAttribute..

  

  

   obj: The object to test the value equality of.

   Returns: true if the value of the given object is equal to that of the current; otherwise,false.
  """
  pass
 def GetHashCode(self):
  """
  GetHashCode(self: CategoryAttribute) -> int

  

   Returns the hash code for this attribute.

   Returns: A 32-bit signed integer hash code.
  """
  pass
 def GetLocalizedString(self,*args):
  """
  GetLocalizedString(self: CategoryAttribute,value: str) -> str

  

   Looks up the localized name of the specified category.

  

   value: The identifer for the category to look up.

   Returns: The localized name of the category,or null if a localized name does not exist.
  """
  pass
 def IsDefaultAttribute(self):
  """
  IsDefaultAttribute(self: CategoryAttribute) -> bool

  

   Determines if this attribute is the default.

   Returns: true if the attribute is the default value for this attribute class; otherwise,false.
  """
  pass
 def __eq__(self,*args):
  """ x.__eq__(y) <==> x==y """
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 @staticmethod
 def __new__(self,category=None):
  """
  __new__(cls: type)

  __new__(cls: type,category: str)
  """
  pass
 def __ne__(self,*args):
  pass
 Category=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the name of the category for the property or event that this attribute is applied to.



Get: Category(self: CategoryAttribute) -> str



"""


 Action=None
 Appearance=None
 Asynchronous=None
 Behavior=None
 Data=None
 Default=None
 Design=None
 DragDrop=None
 Focus=None
 Format=None
 Key=None
 Layout=None
 Mouse=None
 WindowStyle=None




class SocketPermissionAttribute(CodeAccessSecurityAttribute,_Attribute):
 """
 Specifies security actions to control System.Net.Sockets.Socket connections. This class cannot be inherited.

 

 SocketPermissionAttribute(action: SecurityAction)
 """
 def CreatePermission(self):
  """
  CreatePermission(self: SocketPermissionAttribute) -> IPermission

  

   Creates and returns a new instance of the System.Net.SocketPermission class.

   Returns: An instance of the System.Net.SocketPermission class that corresponds to the security 

    declaration.
  """
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 @staticmethod
 def __new__(self,action):
  """ __new__(cls: type,action: SecurityAction) """
  pass
 def __reduce_ex__(self,*args):
  pass
 Access=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the network access method that is allowed by this System.Net.SocketPermissionAttribute.



Get: Access(self: SocketPermissionAttribute) -> str



Set: Access(self: SocketPermissionAttribute)=value

"""

 Host=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the DNS host name or IP address that is specified by this System.Net.SocketPermissionAttribute.



Get: Host(self: SocketPermissionAttribute) -> str



Set: Host(self: SocketPermissionAttribute)=value

"""

 Port=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the port number that is associated with this System.Net.SocketPermissionAttribute.



Get: Port(self: SocketPermissionAttribute) -> str



Set: Port(self: SocketPermissionAttribute)=value

"""

 Transport=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the System.Net.TransportType that is specified by this System.Net.SocketPermissionAttribute.



Get: Transport(self: SocketPermissionAttribute) -> str



Set: Transport(self: SocketPermissionAttribute)=value

"""





class TypeLibVarAttribute(Attribute,_Attribute):
 """
 Contains the System.Runtime.InteropServices.VARFLAGS that were originally imported for this field from the COM type library.

 

 TypeLibVarAttribute(flags: TypeLibVarFlags)

 TypeLibVarAttribute(flags: Int16)
 """
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 @staticmethod
 def __new__(self,flags):
  """
  __new__(cls: type,flags: TypeLibVarFlags)

  __new__(cls: type,flags: Int16)
  """
  pass
 Value=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the System.Runtime.InteropServices.TypeLibVarFlags value for this field.



Get: Value(self: TypeLibVarAttribute) -> TypeLibVarFlags



"""





class DataGridViewCellStyle(object,ICloneable):
 """
 Represents the formatting and style information applied to individual cells within a System.Windows.Forms.DataGridView control.

 

 DataGridViewCellStyle()

 DataGridViewCellStyle(dataGridViewCellStyle: DataGridViewCellStyle)
 """
 def ApplyStyle(self,dataGridViewCellStyle):
  """
  ApplyStyle(self: DataGridViewCellStyle,dataGridViewCellStyle: DataGridViewCellStyle)

   Applies the specified System.Windows.Forms.DataGridViewCellStyle to the current 

    System.Windows.Forms.DataGridViewCellStyle.

  

  

   dataGridViewCellStyle: The System.Windows.Forms.DataGridViewCellStyle to apply to the current 

    System.Windows.Forms.DataGridViewCellStyle.
  """
  pass
 def Clone(self):
  """
  Clone(self: DataGridViewCellStyle) -> DataGridViewCellStyle

  

   Creates an exact copy of this System.Windows.Forms.DataGridViewCellStyle.

   Returns: A System.Windows.Forms.DataGridViewCellStyle that represents an exact copy of this cell style.
  """
  pass
 def Equals(self,o):
  """
  Equals(self: DataGridViewCellStyle,o: object) -> bool

  

   Returns a value indicating whether this instance is equivalent to the specified object.

  

   o: An object to compare with this instance,or null.

   Returns: true if o is a System.Windows.Forms.DataGridViewCellStyle and has the same property values as 

    this instance; otherwise,false.
  """
  pass
 def GetHashCode(self):
  """
  GetHashCode(self: DataGridViewCellStyle) -> int

   Returns: A hash code for the current System.Object.
  """
  pass
 def ToString(self):
  """
  ToString(self: DataGridViewCellStyle) -> str

  

   Returns a string indicating the current property settings of the 

    System.Windows.Forms.DataGridViewCellStyle.

  

   Returns: A string indicating the current property settings of the 

    System.Windows.Forms.DataGridViewCellStyle.
  """
  pass
 def __eq__(self,*args):
  """ x.__eq__(y) <==> x==y """
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 @staticmethod
 def __new__(self,dataGridViewCellStyle=None):
  """
  __new__(cls: type)

  __new__(cls: type,dataGridViewCellStyle: DataGridViewCellStyle)
  """
  pass
 def __ne__(self,*args):
  pass
 def __repr__(self,*args):
  """ __repr__(self: object) -> str """
  pass
 def __str__(self,*args):
  pass
 Alignment=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets a value indicating the position of the cell content within a System.Windows.Forms.DataGridView cell.



Get: Alignment(self: DataGridViewCellStyle) -> DataGridViewContentAlignment



Set: Alignment(self: DataGridViewCellStyle)=value

"""

 BackColor=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the background color of a System.Windows.Forms.DataGridView cell.



Get: BackColor(self: DataGridViewCellStyle) -> Color



Set: BackColor(self: DataGridViewCellStyle)=value

"""

 DataSourceNullValue=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the value saved to the data source when the user enters a null value into a cell.



Get: DataSourceNullValue(self: DataGridViewCellStyle) -> object



Set: DataSourceNullValue(self: DataGridViewCellStyle)=value

"""

 Font=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the font applied to the textual content of a System.Windows.Forms.DataGridView cell.



Get: Font(self: DataGridViewCellStyle) -> Font



Set: Font(self: DataGridViewCellStyle)=value

"""

 ForeColor=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the foreground color of a System.Windows.Forms.DataGridView cell.



Get: ForeColor(self: DataGridViewCellStyle) -> Color



Set: ForeColor(self: DataGridViewCellStyle)=value

"""

 Format=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the format string applied to the textual content of a System.Windows.Forms.DataGridView cell.



Get: Format(self: DataGridViewCellStyle) -> str



Set: Format(self: DataGridViewCellStyle)=value

"""

 FormatProvider=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the object used to provide culture-specific formatting of System.Windows.Forms.DataGridView cell values.



Get: FormatProvider(self: DataGridViewCellStyle) -> IFormatProvider



Set: FormatProvider(self: DataGridViewCellStyle)=value

"""

 IsDataSourceNullValueDefault=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value indicating whether the System.Windows.Forms.DataGridViewCellStyle.DataSourceNullValue property has been set.



Get: IsDataSourceNullValueDefault(self: DataGridViewCellStyle) -> bool



"""

 IsFormatProviderDefault=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value that indicates whether the System.Windows.Forms.DataGridViewCellStyle.FormatProvider property has been set.



Get: IsFormatProviderDefault(self: DataGridViewCellStyle) -> bool



"""

 IsNullValueDefault=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value indicating whether the System.Windows.Forms.DataGridViewCellStyle.NullValue property has been set.



Get: IsNullValueDefault(self: DataGridViewCellStyle) -> bool



"""

 NullValue=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the System.Windows.Forms.DataGridView cell display value corresponding to a cell value of System.DBNull.Value or null.



Get: NullValue(self: DataGridViewCellStyle) -> object



Set: NullValue(self: DataGridViewCellStyle)=value

"""

 Padding=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the space between the edge of a System.Windows.Forms.DataGridViewCell and its content.



Get: Padding(self: DataGridViewCellStyle) -> Padding



Set: Padding(self: DataGridViewCellStyle)=value

"""

 SelectionBackColor=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the background color used by a System.Windows.Forms.DataGridView cell when it is selected.



Get: SelectionBackColor(self: DataGridViewCellStyle) -> Color



Set: SelectionBackColor(self: DataGridViewCellStyle)=value

"""

 SelectionForeColor=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the foreground color used by a System.Windows.Forms.DataGridView cell when it is selected.



Get: SelectionForeColor(self: DataGridViewCellStyle) -> Color



Set: SelectionForeColor(self: DataGridViewCellStyle)=value

"""

 Tag=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets an object that contains additional data related to the System.Windows.Forms.DataGridViewCellStyle.



Get: Tag(self: DataGridViewCellStyle) -> object



Set: Tag(self: DataGridViewCellStyle)=value

"""

 WrapMode=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets a value indicating whether textual content in a System.Windows.Forms.DataGridView cell is wrapped to subsequent lines or truncated when it is too long to fit on a single line.



Get: WrapMode(self: DataGridViewCellStyle) -> DataGridViewTriState



Set: WrapMode(self: DataGridViewCellStyle)=value

"""





"""Wrapper for YamlDotNet."""
from collections import OrderedDict
import codecs

from pyrevit.framework import StringReader, KeyValuePair
from pyrevit.labs import libyaml
from pyrevit.compat import PY3, PY2


def _convert_yamldotnet_to_dict(ynode, level=0):
    if isinstance(ynode, libyaml.RepresentationModel.YamlMappingNode):
        d = OrderedDict()
        for child in ynode.Children:
            if isinstance(child, KeyValuePair[libyaml.RepresentationModel.YamlNode, libyaml.RepresentationModel.YamlNode]):
                d[child.Key.Value] = _convert_yamldotnet_to_dict(child.Value, level=level+1)
        return d
    elif isinstance(ynode, libyaml.RepresentationModel.YamlSequenceNode):
        v = []
        for child in ynode.Children:
            v.append(_convert_yamldotnet_to_dict(child, level=level+1))
        return v
    elif isinstance(ynode, libyaml.RepresentationModel.YamlScalarNode):
        return ynode.Value


def load(yaml_file):
    """Load Yaml file into YamlDotNet object.

    Args:
        yaml_file (str): file path to yaml file

    Returns:
        (YamlDotNet.RepresentationModel.YamlMappingNode): yaml node
    """
    if PY3:
        with open(yaml_file, 'r', encoding="utf8") as yamlfile:
            yamlstr = libyaml.RepresentationModel.YamlStream()
            yamldata = yamlfile.read()
            yamlstr.Load(StringReader(yamldata))
            if yamlstr.Documents.Count >= 1:
                return yamlstr.Documents[0].RootNode
    else:
        with open(yaml_file, 'r') as yamlfile:
            yamlstr = libyaml.RepresentationModel.YamlStream()
            yamldata = yamlfile.read().decode('utf-8')
            yamlstr.Load(StringReader(yamldata))
            if yamlstr.Documents.Count >= 1:
                return yamlstr.Documents[0].RootNode


def load_as_dict(yaml_file):
    """Load Yaml file into python dict object.

    Args:
        yaml_file (str): file path to yaml file

    Returns:
        (dict[str, Any]): dictionary representing yaml data
    """
    return _convert_yamldotnet_to_dict(load(yaml_file))


def dump_dict(dict_object, yaml_file):
    """Save YamlDotNet object to Yaml file.

    Args:
        dict_object (dict): dict object to be serialized into yaml
        yaml_file (str): file path to yaml file
    """
    ybuilder = libyaml.Serialization.SerializerBuilder().Build()
    serialized_yaml = ybuilder.Serialize(dict_object)
    with codecs.open(yaml_file, 'w', 'utf-8') as yamlfile:
        yamlfile.write(serialized_yaml.replace('\r\n', '\n'))



class DataGridViewCellCancelEventArgs(CancelEventArgs):
 """
 Provides data for System.Windows.Forms.DataGridView.CellBeginEdit and System.Windows.Forms.DataGridView.RowValidating events.

 

 DataGridViewCellCancelEventArgs(columnIndex: int,rowIndex: int)
 """
 @staticmethod
 def __new__(self,columnIndex,rowIndex):
  """ __new__(cls: type,columnIndex: int,rowIndex: int) """
  pass
 ColumnIndex=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the column index of the cell that the event occurs for.



Get: ColumnIndex(self: DataGridViewCellCancelEventArgs) -> int



"""

 RowIndex=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the row index of the cell that the event occurs for.



Get: RowIndex(self: DataGridViewCellCancelEventArgs) -> int



"""


